/* *********************************************************************************************************************
      ____         ____
     /    \____/  /    \          Written by: .. Jay Brockway
     \/     |      |    \         Date: ........ 2021/01/31
            |      |____/         Revised: ..... 2021/07/24
            |      |     \
            |      |      \
       ____ |      |_     /
      /    \| #   /  \___/ #
********************************************************************************************************************* */
#ifndef JB_TASK_PUBLISHER_SUBSCRIBER_H
#define JB_TASK_PUBLISHER_SUBSCRIBER_H

namespace JB
{
    /* *****************************************************************************************************************
    A Subscriber registers a delegate to be called when the publisher publishes a notification.
    ***************************************************************************************************************** */
    class Subscriber
    {
        /* == Class Members ========================================================================================= */
        public: void (*subFunc)();          // Subscriber function to call.
        public: Subscriber* next;           // Next subscriber in the linked list.


	    /* == Class Constructors ==================================================================================== */
        /* -------------------------------------------------------------------------------------------------------------
        Don't allow instantiation of subscribers without parameters.
        ------------------------------------------------------------------------------------------------------------- */
    	private: Subscriber();              

        /* -------------------------------------------------------------------------------------------------------------
        Declare a new Subscriber instance.  Specify the function to be called.
        ------------------------------------------------------------------------------------------------------------- */
        public: Subscriber(void (*subFunc)());
    };
}

 
namespace JB
{
    /* *****************************************************************************************************************
    A Publisher publishes a notification when an event occurs to notify subscribers of the event.
    Design:
    *) Allows programs to be written uncoupled; without nested function call.
    *) Eliminates the need to do polling (such as reading a sensor) and thus maximizes CPU usage.
    *) Subscribers are added and removed at runtime allowing a program to be more dynamic.
    *) Used within a multi-threaded OS, the publisher could be modified to run each subscriber handler in a separate 
       thread.  Separate threads would allow the subscriber to perform CPU intensive tasks without blocking the publisher.
    *) Increases readability, maintainability, extensibility and performance.
    *) Minimal design makes it simple to understand and use.
    Usage:
    // Gobal Variables
    Publisher* publisherOne;
    void setup()
    {
        publisherOne = new Publisher();
        publisherOne->Add(new Subscriber(SubscriberHandler));
    }
    void PublisherFunction()
    {
        publisherOne->Notify();
    }
    void SubscriberHandler() {}
    ***************************************************************************************************************** */
    class Publisher
    {
        /* == Class Members ========================================================================================= */
        private: Subscriber* head = NULL;
        private: Subscriber* tail = NULL;


	    /* == Class Constructors ==================================================================================== */
        /* -------------------------------------------------------------------------------------------------------------
        Create a new instance of Publisher.
        ------------------------------------------------------------------------------------------------------------- */
        public: Publisher();


        /* == Class Functions ======================================================================================= */
        /* -------------------------------------------------------------------------------------------------------------
        Add a new subscriber to the subscriber list.
        ------------------------------------------------------------------------------------------------------------- */
        public: void Add(Subscriber* subscriber);

        /* -------------------------------------------------------------------------------------------------------------
        Remove a subscriber from the subscriber list.
        Example: How to add and later remove a subscriber.
            Subscriber *subscriberC = new Subscriber(ButtonC);
            publisherButtonA->Add(subscriberC);
            publisherButtonA->Remove(subscriberC);
        ------------------------------------------------------------------------------------------------------------- */
        public: void Remove(Subscriber* subscriber);

        /* -------------------------------------------------------------------------------------------------------------
        Remove all subscribers from publisher.
        ------------------------------------------------------------------------------------------------------------- */
        public: void Clear();

        /* -------------------------------------------------------------------------------------------------------------
        Notify all the subscribers of the event.
        ------------------------------------------------------------------------------------------------------------- */
        public: void Notify();
    };
}
#endif
