/* *********************************************************************************************************************
      ____         ____
     /    \____/  /    \          Written by: .. Jay Brockway
     \/     |      |    \         Date: ........ 2021/01/31
            |      |____/         Revised: ..... 2021/07/24
            |      |     \
            |      |      \
       ____ |      |_     /
      /    \| #   /  \___/ #
********************************************************************************************************************* */
#ifndef JB_TASK_SCHEDULER_H
#define JB_TASK_SCHEDULER_H


namespace JB
{
    /* *****************************************************************************************************************
        A Task encapsulates a pointer to the function to be called along with the period of time between calls.
    ***************************************************************************************************************** */
    class Task
    {
        /* == Class Members ========================================================================================= */
        public: void (*taskFunc)();         // Task function to call when period expires.
        public: Task* next;                 // Next task in linked list.

        public: int period;                 // Period between calls in milliseconds.
        public: unsigned long nextTime;     // The next time to call the function.


	    /* == Class Constructors ==================================================================================== */
        /* -------------------------------------------------------------------------------------------------------------
        Don't allow instantiation of tasks without parameters.
        ------------------------------------------------------------------------------------------------------------- */
        public: Task();


        /* -------------------------------------------------------------------------------------------------------------
        Declare a new Task instance.  Specify the function to be called and the period of time between calls in 
        milliseconds.
        ------------------------------------------------------------------------------------------------------------- */
        public: Task(void (*taskFunc)(), int period);


        /* == Class Functions ======================================================================================= */
        /* -------------------------------------------------------------------------------------------------------------
        Schedule the next time this task needs to run.
        ------------------------------------------------------------------------------------------------------------- */
        public: void Schedule();
    };
}

 
namespace JB
{
    /* *****************************************************************************************************************
    A Scheduler looks thru a list of tasks for those whose next run time has past and calls their function.
    Design:
    *) Implements nonpreemptive (cooperative) round-robin scheduling.
    *) Eliminates the need to use delay() and thus maximizes CPU usage.
    *) Provides a mechanism to perform CPU intensive tasks outside the interrupt handler to avoid blocking the interrupt.
    *) Tasks are added and removed at runtime allowing a program to be more dynamic.
    *) Increases readability, maintainability, extensibility and performance.
    *) Minimal design makes it simple to understand and use.
    Usage:
    // Global variables
    Scheduler* scheduler = new Scheduler();
    bool isLedOneOn = false;

    setup()
    {
        scheduler->Add(new Task(BlinkLedOne, 500));
    }

    loop()
    {
        scheduler->Run();
    }

    void BlinkLedOne()
    {
        isLedOneOn = !isLedOneOn;
        if (isLedOneOn)
            digitalWrite(LED_1, LOW);
        else
            digitalWrite(LED_1, HIGH);
    }
    ***************************************************************************************************************** */
    class Scheduler
    {
        /* == Class Members ========================================================================================= */
        private: Task* head = NULL;         // First task in linked list.
        private: Task* tail = NULL;         // Last task in linked list.


	    /* == Class Constructors ==================================================================================== */
        /* -------------------------------------------------------------------------------------------------------------
        Create a new instance of Scheduler.
        ------------------------------------------------------------------------------------------------------------- */
        public: Scheduler();


        /* == Class Functions ======================================================================================= */
        /* -------------------------------------------------------------------------------------------------------------
        Add a new task to the scheduler.
        ------------------------------------------------------------------------------------------------------------- */
        public: void Add(Task* task);

        /* -------------------------------------------------------------------------------------------------------------
        Remove the specified task from the scheduler.
        ------------------------------------------------------------------------------------------------------------- */
        public: void Remove(Task* task);

        /* -------------------------------------------------------------------------------------------------------------
        Loop thru the linked list of tasks and call a task's function if it's scheduled time has elapsed.
        ------------------------------------------------------------------------------------------------------------- */
        public: void Run();
    };
}
#endif
